From 066cde03a949b7951206dd2247ee607ab1192ceb Mon Sep 17 00:00:00 2001
From: nkey <michail.nikolaev@gmail.com>
Date: Sun, 24 Nov 2024 14:45:41 +0100
Subject: [PATCH v6 1/4] Specs top reproduce the issues with CREATE INDEX
 CONCURRENTLY and REINDEX CONCURRENTLY in scenarios involving INSERT ON
 CONFLICT DO UPDATE. These tests reproduce different error cases related to
 "duplicate key value violates unique constraint" where this error should not
 occur by design.

* REINDEX CONCURRENTLY and UPSERT with inferred index
* CREATE INDEX CONCURRENTLY and UPSERT with inferred indexes
* REINDEX CONCURRENTLY on partitioned table
* REINDEX CONCURRENTLY with specified constraint name
* CREATE INDEX CONCURRENTLY with predicates

In each of these scenarios, the expected behavior is that the INSERT ON CONFLICT DO UPDATE should handle conflicts gracefully without raising a "duplicate key value violates unique constraint" error. However, due to the concurrent operations on the indexes, this error is encountered.
---
 src/backend/commands/indexcmds.c              |   4 +-
 src/backend/executor/execIndexing.c           |   3 +
 src/backend/executor/nodeModifyTable.c        |   2 +
 src/backend/utils/time/snapmgr.c              |   2 +
 src/test/modules/injection_points/Makefile    |   7 +-
 .../expected/index_concurrently_upsert.out    | 106 ++++++
 .../index_concurrently_upsert_predicate.out   | 106 ++++++
 .../expected/reindex_concurrently_upsert.out  | 316 ++++++++++++++++++
 ...ndex_concurrently_upsert_on_constraint.out | 316 ++++++++++++++++++
 ...eindex_concurrently_upsert_partitioned.out | 316 ++++++++++++++++++
 src/test/modules/injection_points/meson.build |  11 +
 .../specs/index_concurrently_upsert.spec      |  96 ++++++
 .../index_concurrently_upsert_predicate.spec  |  98 ++++++
 .../specs/reindex_concurrently_upsert.spec    | 114 +++++++
 ...dex_concurrently_upsert_on_constraint.spec | 115 +++++++
 ...index_concurrently_upsert_partitioned.spec | 116 +++++++
 16 files changed, 1726 insertions(+), 2 deletions(-)
 create mode 100644 src/test/modules/injection_points/expected/index_concurrently_upsert.out
 create mode 100644 src/test/modules/injection_points/expected/index_concurrently_upsert_predicate.out
 create mode 100644 src/test/modules/injection_points/expected/reindex_concurrently_upsert.out
 create mode 100644 src/test/modules/injection_points/expected/reindex_concurrently_upsert_on_constraint.out
 create mode 100644 src/test/modules/injection_points/expected/reindex_concurrently_upsert_partitioned.out
 create mode 100644 src/test/modules/injection_points/specs/index_concurrently_upsert.spec
 create mode 100644 src/test/modules/injection_points/specs/index_concurrently_upsert_predicate.spec
 create mode 100644 src/test/modules/injection_points/specs/reindex_concurrently_upsert.spec
 create mode 100644 src/test/modules/injection_points/specs/reindex_concurrently_upsert_on_constraint.spec
 create mode 100644 src/test/modules/injection_points/specs/reindex_concurrently_upsert_partitioned.spec

diff --git a/src/backend/commands/indexcmds.c b/src/backend/commands/indexcmds.c
index d6e23caef17..0ff498c4e14 100644
--- a/src/backend/commands/indexcmds.c
+++ b/src/backend/commands/indexcmds.c
@@ -1766,6 +1766,7 @@ DefineIndex(Oid tableId,
 	 * before the reference snap was taken, we have to wait out any
 	 * transactions that might have older snapshots.
 	 */
+	INJECTION_POINT("define_index_before_set_valid");
 	pgstat_progress_update_param(PROGRESS_CREATEIDX_PHASE,
 								 PROGRESS_CREATEIDX_PHASE_WAIT_3);
 	WaitForOlderSnapshots(limitXmin, true);
@@ -4206,7 +4207,7 @@ ReindexRelationConcurrently(const ReindexStmt *stmt, Oid relationOid, const Rein
 	 * the same time to make sure we only get constraint violations from the
 	 * indexes with the correct names.
 	 */
-
+	INJECTION_POINT("reindex_relation_concurrently_before_swap");
 	StartTransactionCommand();
 
 	/*
@@ -4285,6 +4286,7 @@ ReindexRelationConcurrently(const ReindexStmt *stmt, Oid relationOid, const Rein
 	 * index_drop() for more details.
 	 */
 
+	INJECTION_POINT("reindex_relation_concurrently_before_set_dead");
 	pgstat_progress_update_param(PROGRESS_CREATEIDX_PHASE,
 								 PROGRESS_CREATEIDX_PHASE_WAIT_4);
 	WaitForLockersMultiple(lockTags, AccessExclusiveLock, true);
diff --git a/src/backend/executor/execIndexing.c b/src/backend/executor/execIndexing.c
index 7c87f012c30..ae11c1dd463 100644
--- a/src/backend/executor/execIndexing.c
+++ b/src/backend/executor/execIndexing.c
@@ -117,6 +117,7 @@
 #include "utils/multirangetypes.h"
 #include "utils/rangetypes.h"
 #include "utils/snapmgr.h"
+#include "utils/injection_point.h"
 
 /* waitMode argument to check_exclusion_or_unique_constraint() */
 typedef enum
@@ -936,6 +937,8 @@ retry:
 	econtext->ecxt_scantuple = save_scantuple;
 
 	ExecDropSingleTupleTableSlot(existing_slot);
+	if (!conflict)
+		INJECTION_POINT("check_exclusion_or_unique_constraint_no_conflict");
 
 	return !conflict;
 }
diff --git a/src/backend/executor/nodeModifyTable.c b/src/backend/executor/nodeModifyTable.c
index 1af8c9caf6c..8a1a085b106 100644
--- a/src/backend/executor/nodeModifyTable.c
+++ b/src/backend/executor/nodeModifyTable.c
@@ -69,6 +69,7 @@
 #include "utils/datum.h"
 #include "utils/rel.h"
 #include "utils/snapmgr.h"
+#include "utils/injection_point.h"
 
 
 typedef struct MTTargetRelLookup
@@ -1087,6 +1088,7 @@ ExecInsert(ModifyTableContext *context,
 					return NULL;
 				}
 			}
+			INJECTION_POINT("exec_insert_before_insert_speculative");
 
 			/*
 			 * Before we start insertion proper, acquire our "speculative
diff --git a/src/backend/utils/time/snapmgr.c b/src/backend/utils/time/snapmgr.c
index 8f1508b1ee2..3d018c3a1e8 100644
--- a/src/backend/utils/time/snapmgr.c
+++ b/src/backend/utils/time/snapmgr.c
@@ -64,6 +64,7 @@
 #include "utils/resowner.h"
 #include "utils/snapmgr.h"
 #include "utils/syscache.h"
+#include "utils/injection_point.h"
 
 
 /*
@@ -388,6 +389,7 @@ InvalidateCatalogSnapshot(void)
 		pairingheap_remove(&RegisteredSnapshots, &CatalogSnapshot->ph_node);
 		CatalogSnapshot = NULL;
 		SnapshotResetXmin();
+		INJECTION_POINT("invalidate_catalog_snapshot_end");
 	}
 }
 
diff --git a/src/test/modules/injection_points/Makefile b/src/test/modules/injection_points/Makefile
index 0753a9df58c..f8f86e8f3b6 100644
--- a/src/test/modules/injection_points/Makefile
+++ b/src/test/modules/injection_points/Makefile
@@ -13,7 +13,12 @@ PGFILEDESC = "injection_points - facility for injection points"
 REGRESS = injection_points reindex_conc
 REGRESS_OPTS = --dlpath=$(top_builddir)/src/test/regress
 
-ISOLATION = basic inplace
+ISOLATION = basic inplace \
+			reindex_concurrently_upsert \
+			index_concurrently_upsert \
+			reindex_concurrently_upsert_partitioned \
+			reindex_concurrently_upsert_on_constraint \
+			index_concurrently_upsert_predicate
 
 TAP_TESTS = 1
 
diff --git a/src/test/modules/injection_points/expected/index_concurrently_upsert.out b/src/test/modules/injection_points/expected/index_concurrently_upsert.out
new file mode 100644
index 00000000000..4a65abbb55e
--- /dev/null
+++ b/src/test/modules/injection_points/expected/index_concurrently_upsert.out
@@ -0,0 +1,106 @@
+Parsed test spec with 4 sessions
+
+starting permutation: s3_start_create_index s1_start_upsert s4_wakeup_define_index_before_set_valid s2_start_upsert s4_wakeup_s1_from_invalidate_catalog_snapshot s4_wakeup_s2 s4_wakeup_s1
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+step s3_start_create_index: CREATE UNIQUE INDEX CONCURRENTLY tbl_pkey_duplicate ON test.tbl(i); <waiting ...>
+step s1_start_upsert: 
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict(i) do update set updated_at = now();
+		RAISE NOTICE 'upsert 1 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+ <waiting ...>
+step s4_wakeup_define_index_before_set_valid: 
+	SELECT injection_points_detach('define_index_before_set_valid');
+	SELECT injection_points_wakeup('define_index_before_set_valid');
+	call raise_notice('wakeup before set valid done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup before set valid done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s2_start_upsert: 
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict(i) do update set updated_at = now();
+		RAISE NOTICE 'upsert 2 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+ <waiting ...>
+step s4_wakeup_s1_from_invalidate_catalog_snapshot: 
+	SELECT injection_points_detach('invalidate_catalog_snapshot_end');
+	SELECT injection_points_wakeup('invalidate_catalog_snapshot_end');
+	call raise_notice('wakeup 1 done again');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 1 done again
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s4_wakeup_s2: 
+	SELECT injection_points_detach('exec_insert_before_insert_speculative');
+	SELECT injection_points_wakeup('exec_insert_before_insert_speculative');
+	call raise_notice('wakeup 2 done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 2 done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+s2: NOTICE:  upsert 2 done
+step s2_start_upsert: <... completed>
+step s4_wakeup_s1: 
+	SELECT injection_points_detach('check_exclusion_or_unique_constraint_no_conflict');
+	SELECT injection_points_wakeup('check_exclusion_or_unique_constraint_no_conflict');
+	call raise_notice('wakeup 1 done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 1 done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s3_start_create_index: <... completed>
+s1: NOTICE:  upsert 1 done
+step s1_start_upsert: <... completed>
diff --git a/src/test/modules/injection_points/expected/index_concurrently_upsert_predicate.out b/src/test/modules/injection_points/expected/index_concurrently_upsert_predicate.out
new file mode 100644
index 00000000000..7c7ddc7a7f6
--- /dev/null
+++ b/src/test/modules/injection_points/expected/index_concurrently_upsert_predicate.out
@@ -0,0 +1,106 @@
+Parsed test spec with 4 sessions
+
+starting permutation: s3_start_create_index s1_start_upsert s4_wakeup_define_index_before_set_valid s2_start_upsert s4_wakeup_s1_from_invalidate_catalog_snapshot s4_wakeup_s2 s4_wakeup_s1
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+step s3_start_create_index: CREATE UNIQUE INDEX CONCURRENTLY tbl_pkey_special_duplicate ON test.tbl(abs(i)) WHERE i < 10000; <waiting ...>
+step s1_start_upsert: 
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict(abs(i)) where i < 100 do update set updated_at = now();
+		RAISE NOTICE 'upsert 1 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+ <waiting ...>
+step s4_wakeup_define_index_before_set_valid: 
+	SELECT injection_points_detach('define_index_before_set_valid');
+	SELECT injection_points_wakeup('define_index_before_set_valid');
+	call raise_notice('wakeup before set valid done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup before set valid done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s2_start_upsert: 
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now())  on conflict(abs(i)) where i < 100 do update set updated_at = now();
+		RAISE NOTICE 'upsert 2 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+ <waiting ...>
+step s4_wakeup_s1_from_invalidate_catalog_snapshot: 
+	SELECT injection_points_detach('invalidate_catalog_snapshot_end');
+	SELECT injection_points_wakeup('invalidate_catalog_snapshot_end');
+	call raise_notice('wakeup 1 done again');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 1 done again
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s4_wakeup_s2: 
+	SELECT injection_points_detach('exec_insert_before_insert_speculative');
+	SELECT injection_points_wakeup('exec_insert_before_insert_speculative');
+	call raise_notice('wakeup 2 done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 2 done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+s2: NOTICE:  upsert 2 done
+step s2_start_upsert: <... completed>
+step s4_wakeup_s1: 
+	SELECT injection_points_detach('check_exclusion_or_unique_constraint_no_conflict');
+	SELECT injection_points_wakeup('check_exclusion_or_unique_constraint_no_conflict');
+	call raise_notice('wakeup 1 done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 1 done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s3_start_create_index: <... completed>
+s1: NOTICE:  upsert 1 done
+step s1_start_upsert: <... completed>
diff --git a/src/test/modules/injection_points/expected/reindex_concurrently_upsert.out b/src/test/modules/injection_points/expected/reindex_concurrently_upsert.out
new file mode 100644
index 00000000000..cb597620fe3
--- /dev/null
+++ b/src/test/modules/injection_points/expected/reindex_concurrently_upsert.out
@@ -0,0 +1,316 @@
+Parsed test spec with 4 sessions
+
+starting permutation: s3_start_reindex s1_start_upsert s4_wakeup_to_swap s2_start_upsert s4_wakeup_s1 s4_wakeup_s2 s4_wakeup_to_set_dead
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+step s3_start_reindex: REINDEX INDEX CONCURRENTLY test.tbl_pkey; <waiting ...>
+step s1_start_upsert: 
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict(i) do update set updated_at = now();
+		RAISE NOTICE 'upsert 1 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+ <waiting ...>
+step s4_wakeup_to_swap: 
+	SELECT injection_points_detach('reindex_relation_concurrently_before_swap');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_swap');
+	call raise_notice('wakeup before swap done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup before swap done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s2_start_upsert: 
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict(i) do update set updated_at = now();
+		RAISE NOTICE 'upsert 2 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+ <waiting ...>
+step s4_wakeup_s1: 
+	SELECT injection_points_detach('check_exclusion_or_unique_constraint_no_conflict');
+	SELECT injection_points_wakeup('check_exclusion_or_unique_constraint_no_conflict');
+	call raise_notice('wakeup 1 done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 1 done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+s1: NOTICE:  upsert 1 done
+step s1_start_upsert: <... completed>
+step s4_wakeup_s2: 
+	SELECT injection_points_detach('exec_insert_before_insert_speculative');
+	SELECT injection_points_wakeup('exec_insert_before_insert_speculative');
+	call raise_notice('wakeup 2 done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 2 done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+s2: NOTICE:  upsert 2 done
+step s2_start_upsert: <... completed>
+step s4_wakeup_to_set_dead: 
+	SELECT injection_points_detach('reindex_relation_concurrently_before_set_dead');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_set_dead');
+	call raise_notice('wakeup before set dead done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup before set dead done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s3_start_reindex: <... completed>
+
+starting permutation: s3_start_reindex s2_start_upsert s4_wakeup_to_swap s1_start_upsert s4_wakeup_s1 s4_wakeup_s2 s4_wakeup_to_set_dead
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+step s3_start_reindex: REINDEX INDEX CONCURRENTLY test.tbl_pkey; <waiting ...>
+step s2_start_upsert: 
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict(i) do update set updated_at = now();
+		RAISE NOTICE 'upsert 2 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+ <waiting ...>
+step s4_wakeup_to_swap: 
+	SELECT injection_points_detach('reindex_relation_concurrently_before_swap');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_swap');
+	call raise_notice('wakeup before swap done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup before swap done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s1_start_upsert: 
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict(i) do update set updated_at = now();
+		RAISE NOTICE 'upsert 1 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+ <waiting ...>
+step s4_wakeup_s1: 
+	SELECT injection_points_detach('check_exclusion_or_unique_constraint_no_conflict');
+	SELECT injection_points_wakeup('check_exclusion_or_unique_constraint_no_conflict');
+	call raise_notice('wakeup 1 done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 1 done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+s1: NOTICE:  upsert 1 done
+step s1_start_upsert: <... completed>
+step s4_wakeup_s2: 
+	SELECT injection_points_detach('exec_insert_before_insert_speculative');
+	SELECT injection_points_wakeup('exec_insert_before_insert_speculative');
+	call raise_notice('wakeup 2 done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 2 done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+s2: NOTICE:  upsert 2 done
+step s2_start_upsert: <... completed>
+step s4_wakeup_to_set_dead: 
+	SELECT injection_points_detach('reindex_relation_concurrently_before_set_dead');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_set_dead');
+	call raise_notice('wakeup before set dead done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup before set dead done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s3_start_reindex: <... completed>
+
+starting permutation: s3_start_reindex s4_wakeup_to_swap s1_start_upsert s2_start_upsert s4_wakeup_s1 s4_wakeup_to_set_dead s4_wakeup_s2
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+step s3_start_reindex: REINDEX INDEX CONCURRENTLY test.tbl_pkey; <waiting ...>
+step s4_wakeup_to_swap: 
+	SELECT injection_points_detach('reindex_relation_concurrently_before_swap');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_swap');
+	call raise_notice('wakeup before swap done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup before swap done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s1_start_upsert: 
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict(i) do update set updated_at = now();
+		RAISE NOTICE 'upsert 1 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+ <waiting ...>
+step s2_start_upsert: 
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict(i) do update set updated_at = now();
+		RAISE NOTICE 'upsert 2 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+ <waiting ...>
+step s4_wakeup_s1: 
+	SELECT injection_points_detach('check_exclusion_or_unique_constraint_no_conflict');
+	SELECT injection_points_wakeup('check_exclusion_or_unique_constraint_no_conflict');
+	call raise_notice('wakeup 1 done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 1 done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+s1: NOTICE:  upsert 1 done
+step s1_start_upsert: <... completed>
+step s4_wakeup_to_set_dead: 
+	SELECT injection_points_detach('reindex_relation_concurrently_before_set_dead');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_set_dead');
+	call raise_notice('wakeup before set dead done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup before set dead done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s4_wakeup_s2: 
+	SELECT injection_points_detach('exec_insert_before_insert_speculative');
+	SELECT injection_points_wakeup('exec_insert_before_insert_speculative');
+	call raise_notice('wakeup 2 done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 2 done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+s2: NOTICE:  upsert 2 done
+step s2_start_upsert: <... completed>
+step s3_start_reindex: <... completed>
diff --git a/src/test/modules/injection_points/expected/reindex_concurrently_upsert_on_constraint.out b/src/test/modules/injection_points/expected/reindex_concurrently_upsert_on_constraint.out
new file mode 100644
index 00000000000..b29caa94e50
--- /dev/null
+++ b/src/test/modules/injection_points/expected/reindex_concurrently_upsert_on_constraint.out
@@ -0,0 +1,316 @@
+Parsed test spec with 4 sessions
+
+starting permutation: s3_start_reindex s1_start_upsert s4_wakeup_to_swap s2_start_upsert s4_wakeup_s1 s4_wakeup_s2 s4_wakeup_to_set_dead
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+step s3_start_reindex: REINDEX INDEX CONCURRENTLY test.tbl_pkey; <waiting ...>
+step s1_start_upsert: 
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict on constraint tbl_pkey do update set updated_at = now();
+		RAISE NOTICE 'upsert 1 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+ <waiting ...>
+step s4_wakeup_to_swap: 
+	SELECT injection_points_detach('reindex_relation_concurrently_before_swap');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_swap');
+	call raise_notice('wakeup before swap done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup before swap done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s2_start_upsert: 
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict on constraint tbl_pkey do update set updated_at = now();
+		RAISE NOTICE 'upsert 2 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+ <waiting ...>
+step s4_wakeup_s1: 
+	SELECT injection_points_detach('check_exclusion_or_unique_constraint_no_conflict');
+	SELECT injection_points_wakeup('check_exclusion_or_unique_constraint_no_conflict');
+	call raise_notice('wakeup 1 done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 1 done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+s1: NOTICE:  upsert 1 done
+step s1_start_upsert: <... completed>
+step s4_wakeup_s2: 
+	SELECT injection_points_detach('exec_insert_before_insert_speculative');
+	SELECT injection_points_wakeup('exec_insert_before_insert_speculative');
+	call raise_notice('wakeup 2 done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 2 done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+s2: NOTICE:  upsert 2 done
+step s2_start_upsert: <... completed>
+step s4_wakeup_to_set_dead: 
+	SELECT injection_points_detach('reindex_relation_concurrently_before_set_dead');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_set_dead');
+	call raise_notice('wakeup before set dead done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup before set dead done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s3_start_reindex: <... completed>
+
+starting permutation: s3_start_reindex s2_start_upsert s4_wakeup_to_swap s1_start_upsert s4_wakeup_s1 s4_wakeup_s2 s4_wakeup_to_set_dead
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+step s3_start_reindex: REINDEX INDEX CONCURRENTLY test.tbl_pkey; <waiting ...>
+step s2_start_upsert: 
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict on constraint tbl_pkey do update set updated_at = now();
+		RAISE NOTICE 'upsert 2 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+ <waiting ...>
+step s4_wakeup_to_swap: 
+	SELECT injection_points_detach('reindex_relation_concurrently_before_swap');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_swap');
+	call raise_notice('wakeup before swap done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup before swap done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s1_start_upsert: 
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict on constraint tbl_pkey do update set updated_at = now();
+		RAISE NOTICE 'upsert 1 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+ <waiting ...>
+step s4_wakeup_s1: 
+	SELECT injection_points_detach('check_exclusion_or_unique_constraint_no_conflict');
+	SELECT injection_points_wakeup('check_exclusion_or_unique_constraint_no_conflict');
+	call raise_notice('wakeup 1 done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 1 done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+s1: NOTICE:  upsert 1 done
+step s1_start_upsert: <... completed>
+step s4_wakeup_s2: 
+	SELECT injection_points_detach('exec_insert_before_insert_speculative');
+	SELECT injection_points_wakeup('exec_insert_before_insert_speculative');
+	call raise_notice('wakeup 2 done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 2 done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+s2: NOTICE:  upsert 2 done
+step s2_start_upsert: <... completed>
+step s4_wakeup_to_set_dead: 
+	SELECT injection_points_detach('reindex_relation_concurrently_before_set_dead');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_set_dead');
+	call raise_notice('wakeup before set dead done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup before set dead done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s3_start_reindex: <... completed>
+
+starting permutation: s3_start_reindex s4_wakeup_to_swap s1_start_upsert s2_start_upsert s4_wakeup_s1 s4_wakeup_to_set_dead s4_wakeup_s2
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+step s3_start_reindex: REINDEX INDEX CONCURRENTLY test.tbl_pkey; <waiting ...>
+step s4_wakeup_to_swap: 
+	SELECT injection_points_detach('reindex_relation_concurrently_before_swap');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_swap');
+	call raise_notice('wakeup before swap done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup before swap done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s1_start_upsert: 
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict on constraint tbl_pkey do update set updated_at = now();
+		RAISE NOTICE 'upsert 1 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+ <waiting ...>
+step s2_start_upsert: 
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict on constraint tbl_pkey do update set updated_at = now();
+		RAISE NOTICE 'upsert 2 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+ <waiting ...>
+step s4_wakeup_s1: 
+	SELECT injection_points_detach('check_exclusion_or_unique_constraint_no_conflict');
+	SELECT injection_points_wakeup('check_exclusion_or_unique_constraint_no_conflict');
+	call raise_notice('wakeup 1 done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 1 done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+s1: NOTICE:  upsert 1 done
+step s1_start_upsert: <... completed>
+step s4_wakeup_to_set_dead: 
+	SELECT injection_points_detach('reindex_relation_concurrently_before_set_dead');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_set_dead');
+	call raise_notice('wakeup before set dead done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup before set dead done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s4_wakeup_s2: 
+	SELECT injection_points_detach('exec_insert_before_insert_speculative');
+	SELECT injection_points_wakeup('exec_insert_before_insert_speculative');
+	call raise_notice('wakeup 2 done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 2 done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+s2: NOTICE:  upsert 2 done
+step s2_start_upsert: <... completed>
+step s3_start_reindex: <... completed>
diff --git a/src/test/modules/injection_points/expected/reindex_concurrently_upsert_partitioned.out b/src/test/modules/injection_points/expected/reindex_concurrently_upsert_partitioned.out
new file mode 100644
index 00000000000..16aebaa2e12
--- /dev/null
+++ b/src/test/modules/injection_points/expected/reindex_concurrently_upsert_partitioned.out
@@ -0,0 +1,316 @@
+Parsed test spec with 4 sessions
+
+starting permutation: s3_start_reindex s1_start_upsert s4_wakeup_to_swap s2_start_upsert s4_wakeup_s1 s4_wakeup_s2 s4_wakeup_to_set_dead
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+step s3_start_reindex: REINDEX INDEX CONCURRENTLY test.tbl_partition_pkey; <waiting ...>
+step s1_start_upsert: 
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict(i) do update set updated_at = now();
+		RAISE NOTICE 'upsert 1 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+ <waiting ...>
+step s4_wakeup_to_swap: 
+	SELECT injection_points_detach('reindex_relation_concurrently_before_swap');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_swap');
+	call raise_notice('wakeup before swap done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup before swap done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s2_start_upsert: 
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict(i) do update set updated_at = now();
+		RAISE NOTICE 'upsert 2 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+ <waiting ...>
+step s4_wakeup_s1: 
+	SELECT injection_points_detach('check_exclusion_or_unique_constraint_no_conflict');
+	SELECT injection_points_wakeup('check_exclusion_or_unique_constraint_no_conflict');
+	call raise_notice('wakeup 1 done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 1 done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+s1: NOTICE:  upsert 1 done
+step s1_start_upsert: <... completed>
+step s4_wakeup_s2: 
+	SELECT injection_points_detach('exec_insert_before_insert_speculative');
+	SELECT injection_points_wakeup('exec_insert_before_insert_speculative');
+	call raise_notice('wakeup 2 done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 2 done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+s2: NOTICE:  upsert 2 done
+step s2_start_upsert: <... completed>
+step s4_wakeup_to_set_dead: 
+	SELECT injection_points_detach('reindex_relation_concurrently_before_set_dead');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_set_dead');
+	call raise_notice('wakeup before set dead done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup before set dead done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s3_start_reindex: <... completed>
+
+starting permutation: s3_start_reindex s2_start_upsert s4_wakeup_to_swap s1_start_upsert s4_wakeup_s1 s4_wakeup_s2 s4_wakeup_to_set_dead
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+step s3_start_reindex: REINDEX INDEX CONCURRENTLY test.tbl_partition_pkey; <waiting ...>
+step s2_start_upsert: 
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict(i) do update set updated_at = now();
+		RAISE NOTICE 'upsert 2 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+ <waiting ...>
+step s4_wakeup_to_swap: 
+	SELECT injection_points_detach('reindex_relation_concurrently_before_swap');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_swap');
+	call raise_notice('wakeup before swap done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup before swap done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s1_start_upsert: 
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict(i) do update set updated_at = now();
+		RAISE NOTICE 'upsert 1 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+ <waiting ...>
+step s4_wakeup_s1: 
+	SELECT injection_points_detach('check_exclusion_or_unique_constraint_no_conflict');
+	SELECT injection_points_wakeup('check_exclusion_or_unique_constraint_no_conflict');
+	call raise_notice('wakeup 1 done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 1 done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+s1: NOTICE:  upsert 1 done
+step s1_start_upsert: <... completed>
+step s4_wakeup_s2: 
+	SELECT injection_points_detach('exec_insert_before_insert_speculative');
+	SELECT injection_points_wakeup('exec_insert_before_insert_speculative');
+	call raise_notice('wakeup 2 done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 2 done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+s2: NOTICE:  upsert 2 done
+step s2_start_upsert: <... completed>
+step s4_wakeup_to_set_dead: 
+	SELECT injection_points_detach('reindex_relation_concurrently_before_set_dead');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_set_dead');
+	call raise_notice('wakeup before set dead done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup before set dead done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s3_start_reindex: <... completed>
+
+starting permutation: s3_start_reindex s4_wakeup_to_swap s1_start_upsert s2_start_upsert s4_wakeup_s1 s4_wakeup_to_set_dead s4_wakeup_s2
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+injection_points_attach
+-----------------------
+                       
+(1 row)
+
+step s3_start_reindex: REINDEX INDEX CONCURRENTLY test.tbl_partition_pkey; <waiting ...>
+step s4_wakeup_to_swap: 
+	SELECT injection_points_detach('reindex_relation_concurrently_before_swap');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_swap');
+	call raise_notice('wakeup before swap done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup before swap done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s1_start_upsert: 
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict(i) do update set updated_at = now();
+		RAISE NOTICE 'upsert 1 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+ <waiting ...>
+step s2_start_upsert: 
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict(i) do update set updated_at = now();
+		RAISE NOTICE 'upsert 2 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+ <waiting ...>
+step s4_wakeup_s1: 
+	SELECT injection_points_detach('check_exclusion_or_unique_constraint_no_conflict');
+	SELECT injection_points_wakeup('check_exclusion_or_unique_constraint_no_conflict');
+	call raise_notice('wakeup 1 done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 1 done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+s1: NOTICE:  upsert 1 done
+step s1_start_upsert: <... completed>
+step s4_wakeup_to_set_dead: 
+	SELECT injection_points_detach('reindex_relation_concurrently_before_set_dead');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_set_dead');
+	call raise_notice('wakeup before set dead done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup before set dead done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+step s4_wakeup_s2: 
+	SELECT injection_points_detach('exec_insert_before_insert_speculative');
+	SELECT injection_points_wakeup('exec_insert_before_insert_speculative');
+	call raise_notice('wakeup 2 done');
+
+injection_points_detach
+-----------------------
+                       
+(1 row)
+
+s4: NOTICE:  wakeup 2 done
+injection_points_wakeup
+-----------------------
+                       
+(1 row)
+
+s2: NOTICE:  upsert 2 done
+step s2_start_upsert: <... completed>
+step s3_start_reindex: <... completed>
diff --git a/src/test/modules/injection_points/meson.build b/src/test/modules/injection_points/meson.build
index 989b4db226b..428f1219349 100644
--- a/src/test/modules/injection_points/meson.build
+++ b/src/test/modules/injection_points/meson.build
@@ -44,7 +44,16 @@ tests += {
     'specs': [
       'basic',
       'inplace',
+      'reindex_concurrently_upsert',
+      'index_concurrently_upsert',
+      'reindex_concurrently_upsert_partitioned',
+      'reindex_concurrently_upsert_on_constraint',
+      'index_concurrently_upsert_predicate',
     ],
+    # The injection points are cluster-wide, so disable installcheck
+    'runningcheck': false,
+    # We waiting for all snapshots, so, avoid parallel test executions
+    'runningcheck-parallel': false,
   },
   'tap': {
     'env': {
@@ -53,5 +62,7 @@ tests += {
     'tests': [
       't/001_stats.pl',
     ],
+    # The injection points are cluster-wide, so disable installcheck
+    'runningcheck': false,
   },
 }
diff --git a/src/test/modules/injection_points/specs/index_concurrently_upsert.spec b/src/test/modules/injection_points/specs/index_concurrently_upsert.spec
new file mode 100644
index 00000000000..a33e3e40b41
--- /dev/null
+++ b/src/test/modules/injection_points/specs/index_concurrently_upsert.spec
@@ -0,0 +1,96 @@
+# Test race conditions involving:
+# - s1: UPSERT a tuple
+# - s2: UPSERT the same tuple
+# - s3: CREATE UNIQUE INDEX CONCURRENTLY
+# - s4: operations with injection points
+
+setup
+{
+	CREATE EXTENSION injection_points;
+	CREATE SCHEMA test;
+	CREATE UNLOGGED TABLE test.tbl(i int primary key, updated_at timestamp);
+	ALTER TABLE test.tbl SET (parallel_workers=0);
+
+	CREATE PROCEDURE raise_notice (s text) LANGUAGE plpgsql AS
+	$$
+	BEGIN
+		RAISE NOTICE '%', s;
+	END;
+	$$;
+}
+
+teardown
+{
+	DROP SCHEMA test CASCADE;
+	DROP EXTENSION injection_points;
+	DROP PROCEDURE raise_notice;
+}
+
+session s1
+setup	{
+	SELECT injection_points_set_local();
+	SELECT injection_points_attach('check_exclusion_or_unique_constraint_no_conflict', 'wait');
+	SELECT injection_points_attach('invalidate_catalog_snapshot_end', 'wait');
+}
+step s1_start_upsert	{
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict(i) do update set updated_at = now();
+		RAISE NOTICE 'upsert 1 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+}
+
+session s2
+setup	{
+	SELECT injection_points_set_local();
+	SELECT injection_points_attach('exec_insert_before_insert_speculative', 'wait');
+}
+step s2_start_upsert	{
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict(i) do update set updated_at = now();
+		RAISE NOTICE 'upsert 2 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+}
+
+session s3
+setup	{
+	SELECT injection_points_set_local();
+	SELECT injection_points_attach('define_index_before_set_valid', 'wait');
+}
+step s3_start_create_index		{ CREATE UNIQUE INDEX CONCURRENTLY tbl_pkey_duplicate ON test.tbl(i); }
+
+session s4
+step s4_wakeup_s1		{
+	SELECT injection_points_detach('check_exclusion_or_unique_constraint_no_conflict');
+	SELECT injection_points_wakeup('check_exclusion_or_unique_constraint_no_conflict');
+	call raise_notice('wakeup 1 done');
+}
+step s4_wakeup_s1_from_invalidate_catalog_snapshot	{
+	SELECT injection_points_detach('invalidate_catalog_snapshot_end');
+	SELECT injection_points_wakeup('invalidate_catalog_snapshot_end');
+	call raise_notice('wakeup 1 done again');
+}
+step s4_wakeup_s2		{
+	SELECT injection_points_detach('exec_insert_before_insert_speculative');
+	SELECT injection_points_wakeup('exec_insert_before_insert_speculative');
+	call raise_notice('wakeup 2 done');
+}
+step s4_wakeup_define_index_before_set_valid	{
+	SELECT injection_points_detach('define_index_before_set_valid');
+	SELECT injection_points_wakeup('define_index_before_set_valid');
+	call raise_notice('wakeup before set valid done');
+}
+
+permutation
+	s3_start_create_index(s2_start_upsert notices 1)
+	s1_start_upsert(s2_start_upsert notices 1)
+	s4_wakeup_define_index_before_set_valid
+	s2_start_upsert
+	s4_wakeup_s1_from_invalidate_catalog_snapshot
+	s4_wakeup_s2
+	s4_wakeup_s1
\ No newline at end of file
diff --git a/src/test/modules/injection_points/specs/index_concurrently_upsert_predicate.spec b/src/test/modules/injection_points/specs/index_concurrently_upsert_predicate.spec
new file mode 100644
index 00000000000..5dc7a71d7d4
--- /dev/null
+++ b/src/test/modules/injection_points/specs/index_concurrently_upsert_predicate.spec
@@ -0,0 +1,98 @@
+# Test race conditions involving:
+# - s1: UPSERT a tuple
+# - s2: UPSERT the same tuple
+# - s3: CREATE UNIQUE INDEX CONCURRENTLY
+# - s4: operations with injection points
+
+setup
+{
+	CREATE EXTENSION injection_points;
+	CREATE SCHEMA test;
+	CREATE UNLOGGED TABLE test.tbl(i int, updated_at timestamp);
+
+	CREATE UNIQUE INDEX tbl_pkey_special ON test.tbl(abs(i)) WHERE i < 1000;
+	ALTER TABLE test.tbl SET (parallel_workers=0);
+
+	CREATE PROCEDURE raise_notice (s text) LANGUAGE plpgsql AS
+	$$
+	BEGIN
+		RAISE NOTICE '%', s;
+	END;
+	$$;
+}
+
+teardown
+{
+	DROP SCHEMA test CASCADE;
+	DROP EXTENSION injection_points;
+	DROP PROCEDURE raise_notice;
+}
+
+session s1
+setup	{
+	SELECT injection_points_set_local();
+	SELECT injection_points_attach('check_exclusion_or_unique_constraint_no_conflict', 'wait');
+	SELECT injection_points_attach('invalidate_catalog_snapshot_end', 'wait');
+}
+step s1_start_upsert	{
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict(abs(i)) where i < 100 do update set updated_at = now();
+		RAISE NOTICE 'upsert 1 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+}
+
+session s2
+setup	{
+	SELECT injection_points_set_local();
+	SELECT injection_points_attach('exec_insert_before_insert_speculative', 'wait');
+}
+step s2_start_upsert	{
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now())  on conflict(abs(i)) where i < 100 do update set updated_at = now();
+		RAISE NOTICE 'upsert 2 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+}
+
+session s3
+setup	{
+	SELECT injection_points_set_local();
+	SELECT injection_points_attach('define_index_before_set_valid', 'wait');
+}
+step s3_start_create_index		{ CREATE UNIQUE INDEX CONCURRENTLY tbl_pkey_special_duplicate ON test.tbl(abs(i)) WHERE i < 10000;}
+
+session s4
+step s4_wakeup_s1		{
+	SELECT injection_points_detach('check_exclusion_or_unique_constraint_no_conflict');
+	SELECT injection_points_wakeup('check_exclusion_or_unique_constraint_no_conflict');
+	call raise_notice('wakeup 1 done');
+}
+step s4_wakeup_s1_from_invalidate_catalog_snapshot	{
+	SELECT injection_points_detach('invalidate_catalog_snapshot_end');
+	SELECT injection_points_wakeup('invalidate_catalog_snapshot_end');
+	call raise_notice('wakeup 1 done again');
+}
+step s4_wakeup_s2		{
+	SELECT injection_points_detach('exec_insert_before_insert_speculative');
+	SELECT injection_points_wakeup('exec_insert_before_insert_speculative');
+	call raise_notice('wakeup 2 done');
+}
+step s4_wakeup_define_index_before_set_valid	{
+	SELECT injection_points_detach('define_index_before_set_valid');
+	SELECT injection_points_wakeup('define_index_before_set_valid');
+	call raise_notice('wakeup before set valid done');
+}
+
+permutation
+	s3_start_create_index(s2_start_upsert notices 1)
+	s1_start_upsert(s2_start_upsert notices 1)
+	s4_wakeup_define_index_before_set_valid
+	s2_start_upsert
+	s4_wakeup_s1_from_invalidate_catalog_snapshot
+	s4_wakeup_s2
+	s4_wakeup_s1
\ No newline at end of file
diff --git a/src/test/modules/injection_points/specs/reindex_concurrently_upsert.spec b/src/test/modules/injection_points/specs/reindex_concurrently_upsert.spec
new file mode 100644
index 00000000000..733a83c91d2
--- /dev/null
+++ b/src/test/modules/injection_points/specs/reindex_concurrently_upsert.spec
@@ -0,0 +1,114 @@
+# Test race conditions involving:
+# - s1: UPSERT a tuple
+# - s2: UPSERT the same tuple
+# - s3: REINDEX concurrent primary key index
+# - s4: operations with injection points
+
+setup
+{
+	CREATE EXTENSION injection_points;
+	CREATE SCHEMA test;
+	CREATE UNLOGGED TABLE test.tbl(i int primary key, updated_at timestamp);
+	ALTER TABLE test.tbl SET (parallel_workers=0);
+
+	CREATE PROCEDURE raise_notice (s text) LANGUAGE plpgsql AS
+	$$
+	BEGIN
+		RAISE NOTICE '%', s;
+	END;
+	$$;
+}
+
+teardown
+{
+	DROP SCHEMA test CASCADE;
+	DROP EXTENSION injection_points;
+	DROP PROCEDURE raise_notice;
+}
+
+session s1
+setup	{
+	SELECT injection_points_set_local();
+	SELECT injection_points_attach('check_exclusion_or_unique_constraint_no_conflict', 'wait');
+}
+step s1_start_upsert	{
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict(i) do update set updated_at = now();
+		RAISE NOTICE 'upsert 1 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+}
+
+session s2
+setup	{
+	SELECT injection_points_set_local();
+	SELECT injection_points_attach('exec_insert_before_insert_speculative', 'wait');
+}
+step s2_start_upsert	{
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict(i) do update set updated_at = now();
+		RAISE NOTICE 'upsert 2 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+}
+
+session s3
+setup	{
+	SELECT injection_points_set_local();
+	SELECT injection_points_attach('reindex_relation_concurrently_before_set_dead', 'wait');
+	SELECT injection_points_attach('reindex_relation_concurrently_before_swap', 'wait');
+}
+step s3_start_reindex			{ REINDEX INDEX CONCURRENTLY test.tbl_pkey; }
+
+session s4
+step s4_wakeup_to_swap		{
+	SELECT injection_points_detach('reindex_relation_concurrently_before_swap');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_swap');
+	call raise_notice('wakeup before swap done');
+}
+step s4_wakeup_s1		{
+	SELECT injection_points_detach('check_exclusion_or_unique_constraint_no_conflict');
+	SELECT injection_points_wakeup('check_exclusion_or_unique_constraint_no_conflict');
+	call raise_notice('wakeup 1 done');
+}
+step s4_wakeup_s2		{
+	SELECT injection_points_detach('exec_insert_before_insert_speculative');
+	SELECT injection_points_wakeup('exec_insert_before_insert_speculative');
+	call raise_notice('wakeup 2 done');
+}
+step s4_wakeup_to_set_dead		{
+	SELECT injection_points_detach('reindex_relation_concurrently_before_set_dead');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_set_dead');
+	call raise_notice('wakeup before set dead done');
+}
+
+permutation
+	s3_start_reindex(s2_start_upsert notices 1)
+	s1_start_upsert(s4_wakeup_s2 notices 1)
+	s4_wakeup_to_swap
+	s2_start_upsert(s1_start_upsert notices 1)
+	s4_wakeup_s1
+	s4_wakeup_s2
+	s4_wakeup_to_set_dead
+
+permutation
+	s3_start_reindex(s1_start_upsert notices 1)
+	s2_start_upsert(s1_start_upsert notices 1)
+	s4_wakeup_to_swap
+	s1_start_upsert
+	s4_wakeup_s1
+	s4_wakeup_s2
+	s4_wakeup_to_set_dead
+
+permutation
+	s3_start_reindex(s2_start_upsert notices 1)
+	s4_wakeup_to_swap
+	s1_start_upsert
+	s2_start_upsert(s1_start_upsert notices 1)
+	s4_wakeup_s1
+	s4_wakeup_to_set_dead
+	s4_wakeup_s2
\ No newline at end of file
diff --git a/src/test/modules/injection_points/specs/reindex_concurrently_upsert_on_constraint.spec b/src/test/modules/injection_points/specs/reindex_concurrently_upsert_on_constraint.spec
new file mode 100644
index 00000000000..7181fee8031
--- /dev/null
+++ b/src/test/modules/injection_points/specs/reindex_concurrently_upsert_on_constraint.spec
@@ -0,0 +1,115 @@
+# Test race conditions involving:
+# - s1: UPSERT a tuple
+# - s2: UPSERT the same tuple
+# - s3: REINDEX concurrent primary key index
+# - s4: operations with injection points
+
+setup
+{
+	CREATE EXTENSION injection_points;
+	CREATE SCHEMA test;
+	CREATE UNLOGGED TABLE test.tbl(i int primary key, updated_at timestamp);
+	ALTER TABLE test.tbl SET (parallel_workers=0);
+
+	CREATE PROCEDURE raise_notice (s text) LANGUAGE plpgsql AS
+	$$
+	BEGIN
+		RAISE NOTICE '%', s;
+	END;
+	$$;
+}
+
+teardown
+{
+	DROP SCHEMA test CASCADE;
+	DROP EXTENSION injection_points;
+	DROP PROCEDURE raise_notice;
+}
+
+session s1
+setup	{
+	SELECT injection_points_set_local();
+	SELECT injection_points_attach('check_exclusion_or_unique_constraint_no_conflict', 'wait');
+}
+step s1_start_upsert	{
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict on constraint tbl_pkey do update set updated_at = now();
+		RAISE NOTICE 'upsert 1 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+}
+
+session s2
+setup	{
+	SELECT injection_points_set_local();
+	SELECT injection_points_attach('exec_insert_before_insert_speculative', 'wait');
+}
+step s2_start_upsert	{
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict on constraint tbl_pkey do update set updated_at = now();
+		RAISE NOTICE 'upsert 2 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+}
+
+
+session s3
+setup	{
+	SELECT injection_points_set_local();
+	SELECT injection_points_attach('reindex_relation_concurrently_before_set_dead', 'wait');
+	SELECT injection_points_attach('reindex_relation_concurrently_before_swap', 'wait');
+}
+step s3_start_reindex			{ REINDEX INDEX CONCURRENTLY test.tbl_pkey; }
+
+session s4
+step s4_wakeup_to_swap		{
+	SELECT injection_points_detach('reindex_relation_concurrently_before_swap');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_swap');
+	call raise_notice('wakeup before swap done');
+}
+step s4_wakeup_s1		{
+	SELECT injection_points_detach('check_exclusion_or_unique_constraint_no_conflict');
+	SELECT injection_points_wakeup('check_exclusion_or_unique_constraint_no_conflict');
+	call raise_notice('wakeup 1 done');
+}
+step s4_wakeup_s2		{
+	SELECT injection_points_detach('exec_insert_before_insert_speculative');
+	SELECT injection_points_wakeup('exec_insert_before_insert_speculative');
+	call raise_notice('wakeup 2 done');
+}
+step s4_wakeup_to_set_dead		{
+	SELECT injection_points_detach('reindex_relation_concurrently_before_set_dead');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_set_dead');
+	call raise_notice('wakeup before set dead done');
+}
+
+permutation
+	s3_start_reindex(s2_start_upsert notices 1)
+	s1_start_upsert(s4_wakeup_s2 notices 1)
+	s4_wakeup_to_swap
+	s2_start_upsert(s1_start_upsert notices 1)
+	s4_wakeup_s1
+	s4_wakeup_s2
+	s4_wakeup_to_set_dead
+
+permutation
+	s3_start_reindex(s1_start_upsert notices 1)
+	s2_start_upsert(s1_start_upsert notices 1)
+	s4_wakeup_to_swap
+	s1_start_upsert
+	s4_wakeup_s1
+	s4_wakeup_s2
+	s4_wakeup_to_set_dead
+
+permutation
+	s3_start_reindex(s2_start_upsert notices 1)
+	s4_wakeup_to_swap
+	s1_start_upsert
+	s2_start_upsert(s1_start_upsert notices 1)
+	s4_wakeup_s1
+	s4_wakeup_to_set_dead
+	s4_wakeup_s2
\ No newline at end of file
diff --git a/src/test/modules/injection_points/specs/reindex_concurrently_upsert_partitioned.spec b/src/test/modules/injection_points/specs/reindex_concurrently_upsert_partitioned.spec
new file mode 100644
index 00000000000..d577889e9c9
--- /dev/null
+++ b/src/test/modules/injection_points/specs/reindex_concurrently_upsert_partitioned.spec
@@ -0,0 +1,116 @@
+# Test race conditions involving:
+# - s1: UPSERT a tuple
+# - s2: UPSERT the same tuple
+# - s3: REINDEX concurrent primary key index
+# - s4: operations with injection points
+
+setup
+{
+	CREATE EXTENSION injection_points;
+	CREATE SCHEMA test;
+	CREATE TABLE test.tbl(i int primary key, updated_at timestamp) PARTITION BY RANGE (i);
+	CREATE TABLE test.tbl_partition PARTITION OF test.tbl
+		FOR VALUES FROM (0) TO (10000)
+		WITH (parallel_workers = 0);
+
+	CREATE PROCEDURE raise_notice (s text) LANGUAGE plpgsql AS
+	$$
+	BEGIN
+		RAISE NOTICE '%', s;
+	END;
+	$$;
+}
+
+teardown
+{
+	DROP SCHEMA test CASCADE;
+	DROP EXTENSION injection_points;
+	DROP PROCEDURE raise_notice;
+}
+
+session s1
+setup	{
+	SELECT injection_points_set_local();
+	SELECT injection_points_attach('check_exclusion_or_unique_constraint_no_conflict', 'wait');
+}
+step s1_start_upsert	{
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict(i) do update set updated_at = now();
+		RAISE NOTICE 'upsert 1 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+}
+
+session s2
+setup	{
+	SELECT injection_points_set_local();
+	SELECT injection_points_attach('exec_insert_before_insert_speculative', 'wait');
+}
+step s2_start_upsert	{
+	DO $$
+	BEGIN
+		INSERT INTO test.tbl VALUES(13,now()) on conflict(i) do update set updated_at = now();
+		RAISE NOTICE 'upsert 2 done';
+	EXCEPTION WHEN OTHERS THEN
+		RAISE NOTICE '% %', SQLERRM, SQLSTATE;
+	END; $$
+}
+
+session s3
+setup	{
+	SELECT injection_points_set_local();
+	SELECT injection_points_attach('reindex_relation_concurrently_before_set_dead', 'wait');
+	SELECT injection_points_attach('reindex_relation_concurrently_before_swap', 'wait');
+}
+step s3_start_reindex			{ REINDEX INDEX CONCURRENTLY test.tbl_partition_pkey; }
+
+session s4
+step s4_wakeup_to_swap		{
+	SELECT injection_points_detach('reindex_relation_concurrently_before_swap');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_swap');
+	call raise_notice('wakeup before swap done');
+}
+step s4_wakeup_s1		{
+	SELECT injection_points_detach('check_exclusion_or_unique_constraint_no_conflict');
+	SELECT injection_points_wakeup('check_exclusion_or_unique_constraint_no_conflict');
+	call raise_notice('wakeup 1 done');
+}
+step s4_wakeup_s2		{
+	SELECT injection_points_detach('exec_insert_before_insert_speculative');
+	SELECT injection_points_wakeup('exec_insert_before_insert_speculative');
+	call raise_notice('wakeup 2 done');
+}
+step s4_wakeup_to_set_dead		{
+	SELECT injection_points_detach('reindex_relation_concurrently_before_set_dead');
+	SELECT injection_points_wakeup('reindex_relation_concurrently_before_set_dead');
+	call raise_notice('wakeup before set dead done');
+}
+
+permutation
+	s3_start_reindex(s2_start_upsert notices 1)
+	s1_start_upsert(s4_wakeup_s2 notices 1)
+	s4_wakeup_to_swap
+	s2_start_upsert(s1_start_upsert notices 1)
+	s4_wakeup_s1
+	s4_wakeup_s2
+	s4_wakeup_to_set_dead
+
+permutation
+	s3_start_reindex(s1_start_upsert notices 1)
+	s2_start_upsert(s1_start_upsert notices 1)
+	s4_wakeup_to_swap
+	s1_start_upsert
+	s4_wakeup_s1
+	s4_wakeup_s2
+	s4_wakeup_to_set_dead
+
+permutation
+	s3_start_reindex(s2_start_upsert notices 1)
+	s4_wakeup_to_swap
+	s1_start_upsert
+	s2_start_upsert(s1_start_upsert notices 1)
+	s4_wakeup_s1
+	s4_wakeup_to_set_dead
+	s4_wakeup_s2
\ No newline at end of file
-- 
2.43.0

